/*******************************************************************/
/*                                                                 */
/* Module           : bigint.cc                                    */
/* Description      : Implementation file for class BigInt.        */
/* Author           : Sebastian Boenisch                           */
/* Email            : Sebastian.Boenisch@iwr.uni-heidelberg.de     */
/* Copyright        : University of Heidelberg                     */
/* Created on       : (see RCS entry below)                        */
/* $Id: bigint.cc,v 1.8 2002/03/05 17:45:21 boenisch Exp $         */
/* Last modified by :                                              */
/* Last modified on :                                              */
/* Update count     :                                              */
/*                                                                 */
/*******************************************************************/
/*                                                                 */
/*  Date      Name               Changes/Extensions                */
/*  ====      ====               ==================                */
/*                                                                 */
/*******************************************************************/



#include "bigint.hh"
#include "utility.hh"


BigInt::BigInt() : sign(positive), wordList() {
}


BigInt::BigInt(int val) {
  sign = (val >= 0) ? positive : negative;
  val  = abs(val);

  for (int bitCount  = 0; 
           bitCount  < BITS_PER_INT; 
           bitCount += BITS_PER_WORD ) {

    Word w = val & WORD_MAX;
    val  >>= BITS_PER_WORD ;
    
    wordList.push_back(w);
  }

  removeLeadingZeros();
  
} // inline BigInt::BigInt(int val)


BigInt::BigInt(const list<Word> initList) 
  : sign(positive),wordList(initList) {
}


BigInt::~BigInt() {
  // Nothing to do. Standard destructor is enough.
}


const BigInt& BigInt::negate() const {
  sign = ( (sign==positive) ? negative : positive); 
  return *this;
}


BigInt BigInt::operator-() const {
  BigInt temp (*this) ; 
  return temp.negate(); 
}


BigInt& BigInt::operator++() {
  return (*this = *this+1); 
}


BigInt& BigInt::operator++(int) { 
  return (*this = *this+1); 
}


BigInt& BigInt::operator--() { 
  return (*this = *this-1); 
}


BigInt& BigInt::operator--(int) {
  return (*this = *this-1); 
}


unsigned long BigInt::size() const {
   return wordList.size(); 
}


SignType BigInt::getSign() const {
  return sign; 
}


float BigInt::toFloat() const {
  list<Word>::const_reverse_iterator i = wordList.rbegin();
  float sum = *i++;
  for (; i != wordList.rend(); i++) {
    sum = WORD_MAX*sum + *i;
  }

  if (sign == negative) {
    sum = -sum;
  }

  return sum;
} // float BigInt::toFloat()


double BigInt::toDouble() const {
  list<Word>::const_reverse_iterator i = wordList.rbegin();
  double sum = *i++;
  for (; i != wordList.rend(); i++)
    sum = WORD_MAX*sum + *i;
  
  if (sign == negative)
    sum = -sum;

  return sum;
} // double BigInt::toDouble()


bool BigInt::isZero() const { 
  if ( (size()==1) && (front()==0) ) {
    return true;
  }
  else {
    return false;
  }
} // bool BigInt::isZero()


bool BigInt::isEven() const {
  assert(size()>0);
  return ( ! (wordList.front()&1) );
} // bool BigInt::isEven()


bool BigInt::isOdd () const {
  assert(size()>0);
  return ( wordList.front()&1 );
} // bool BigInt::isOdd()


bool BigInt::isPositive() const {
  assert(size()>0);
  return (sign == positive);
} // bool BigInt::isPositive()


bool BigInt::isNegative() const {
  assert(size()>0);
  return (sign == negative);
} // bool BigInt::isNegative()


list<Word>::iterator BigInt::begin() {
  return wordList.begin(); 
}


list<Word>::const_iterator BigInt::begin() const {
  return wordList.begin(); 
}


list<Word>::iterator BigInt::end() { 
  return wordList.end(); 
}


list<Word>::const_iterator BigInt::end() const {
  return wordList.end(); 
}


list<Word>::reverse_iterator BigInt::rbegin() { 
  return wordList.rbegin(); 
}


list<Word>::const_reverse_iterator BigInt::rbegin() const {
  return wordList.rbegin(); 
}


list<Word>::reverse_iterator BigInt::rend() { 
  return wordList.rend(); 
}


list<Word>::const_reverse_iterator BigInt::rend() const {
  return wordList.rend(); 
}


Word& BigInt::front() {
  return wordList.front(); 
}


const
Word& BigInt::front() const { 
  return wordList.front(); 
}


Word& BigInt::back()  { 
  return wordList.back(); 
}


const
Word& BigInt::back() const { 
  return wordList.back(); 
}

  
BigInt::BigInt(const string& initString) {
  BigInt result(0);
  
  string::const_iterator i = initString.begin();
  if (*i == '-') {
    sign = negative;
    i++;
  }
  else if (*i == '+') {
    sign = positive;
    i++;
  }
  else {
    sign = positive;
  }

  for (; i != initString.end(); i++) {
    result = result*(BigInt)10 + BigInt(*i-'0');
  }

  wordList = result.wordList;
} // BigInt::BigInt(const string& initString)


// Full Adder. x + y + carry in -> (sum,carry out)
Word BigInt::wordAdd(Word  x       , 
		     Word  y       , 
		     Word  carryIn , 
		     Word& carryOut ) const {

  DoubleWord temporarySum = x + y + carryIn;
  
  carryOut = 
    (temporarySum & WORD_MAX_PLUS_ONE) >> BITS_PER_WORD;

  return (temporarySum & WORD_MAX);
} // Word BigInt::wordAdd(...


Word BigInt::wordSub(Word  x       , 
		     Word  y       , 
		     Word  carryIn , 
		     Word& carryOut ) const {
  
  long difference = x - (y+carryIn);

  if (difference < 0) {
    difference += WORD_MAX_PLUS_ONE;
    carryOut = 1;
  }
  else {
    carryOut = 0;
  }
  
  return difference;
} // Word BigInt::wordSub(...


// Multiplies two Words.
void BigInt::wordMultiply(Word  x          , 
			  Word  y          , 
			  Word& upperResult,
			  Word& lowerResult ) const {

  DoubleWord temporaryProduct = x*y;

  upperResult = (temporaryProduct >> BITS_PER_WORD) & WORD_MAX;
  lowerResult =  temporaryProduct                   & WORD_MAX;
} // void BigInt::wordMultiply(...


BigInt& BigInt::operator+=(const BigInt& y) {

  BigInt& x = *this;   // alias

  // Special cases

  if (this == &y) {
    *this << 1;
    return x;
  }

  if (x.isZero()) {
    x = y;
    return x;
  }

  if (y.isZero())
    return x;
 

  /* There are 4 possible sign patterns for (x,y):
     1.(>=0 , >=0)   
     2.(>=0 , < 0)   
     3.(< 0 , >=0)    
     4.(<0  , < 0)

     The cases 2.-4. can either be reduced to the first one
     or be interpreted as an subtraction.
  */

  // 2.
  if ( (x >= 0) && (y < 0) ) {
    y.negate();
    x -= y;
    y.negate();

    return x;
  }
  
  // 3. & 4.
  if (x < 0) {

    // 3.
    if (y >= 0) {
      x.negate();
      x -= y;
      x.negate();

      return x;
    }
    // 4.
    else {  
      x.negate();
      y.negate();
      x += y;
      x.negate();
      y.negate();

      return x;
    }
  }
  
  // 1. 
  // Now we can assume: x>=0, y>=0.

  Word carry = 0;
  
  list<Word>::      iterator xIterator =   begin();
  list<Word>::const_iterator yIterator = y.begin();
 
  for (; xIterator != end(); xIterator++) { 

    *xIterator = 
      wordAdd( *xIterator,
	       (yIterator!=y.end() ? *yIterator : 0), 
	       carry , 
	       carry  );
    
    if (yIterator != y.end()) {
      yIterator++;
    }

  } // for (; xIterator != end(); ...

  for (; yIterator != y.end(); yIterator++) {

    wordList.push_back( 
		       wordAdd (0, *yIterator, carry, carry) 
		       );
  }

  if (carry) {
    wordList.push_back(carry);
  }

  return *this;
} // BigInt& BigInt::operator+=(const BigInt& y)


BigInt BigInt::operator+(const BigInt& y) const {
  BigInt temp(*this);
  temp += y;
  return temp;
} // BigInt BigInt::operator+(...


BigInt BigInt::operator-(const BigInt& y) const {
  
  // Alias
  const BigInt& x = *this;  

  // Special cases
  if (x.isZero())
    return -y;

  if (y.isZero())
    return *this;


  // General case.

  BigInt result;
  
  /* There are 4 possible sign patterns for (x,y):
     1.(>=0 , >=0)
     2.(>=0 , < 0)
     3.(< 0 , >=0)  
     4.(< 0 , < 0)

     The cases 2.-4. can either be reduced to the first one
     or be interpreted as an addition.
  */
  
  // 2.
  if ( (x >= 0) && (y < 0) ) {
    y.negate();
    result = x + y;
    y.negate();

    return result;
  }
  
  // 3. & 4.
  if (x < 0) {

    // 3.
    if (y >= 0) {
      x.negate();
      result = x+y;
      result.negate();
      x.negate();

      return result;
    }
    // 4.
    else {  
      x.negate();
      y.negate();
      result = y-x;
      x.negate();
      y.negate();

      return result;
    }
  }


  // 1.
  // Now we can assume: x>=0, y>=0.

  // The case x<y can be reduced to the case x>=y:
  if (x<y) {
    result = y-x;
    result.negate();

    return result;
  }  

  // Now we can additionally assume x>=y.
  
  Word subCarry = 0;
  Word addCarry = 0;
  list<Word>::const_iterator xIterator =   begin();
  list<Word>::const_iterator yIterator = y.begin();
  
  for (; xIterator != end(); xIterator++) {

    // Awful expression !
    result.wordList.push_back(
       wordSub(*xIterator,
	       wordAdd( (yIterator!=y.end()?*yIterator:0),
		        0       , 
		        addCarry,
		        addCarry  
		      ) ,
	       subCarry,
	       subCarry
	       ) 
              );

    if (yIterator != y.end()) {
      yIterator++;
    }

  }
  
  result.removeLeadingZeros();

  return result;
} // BigInt BigInt::operator-(const BigInt& y) const


BigInt& BigInt::operator-=(const BigInt& y) {
  BigInt temp = *this;
  *this = *this - y;
  return  *this;
} // BigInt& BigInt::operator-=(const BigInt& y)

  
int BigInt::getBit(unsigned long bit) const {
  list<Word>::const_iterator i = begin();
  unsigned long wordNumber = bit/BITS_PER_WORD;
  
  for(unsigned long j=0; j<wordNumber; i++,j++);
  
  return getBit(bit,i);
} //  int BigInt::getBit(unsigned long bit) const


int 
BigInt::getBit(unsigned int bit, 
	       list<Word>::const_iterator i) const  {
  return ( (*i) >> bit ) & 1;
} // BigInt::getBit(...


SignType BigInt::multDivSignRule(SignType sign1, SignType sign2) const {
  
  if ( (sign1==positive) && (sign2==positive) )
    return positive;
  if ( (sign1==negative) && (sign2== negative) )
    return positive;
  
  return negative;
} // SignType BigInt::multDivSignRule(...


ComparisonType BigInt::compare(const BigInt& x) const {
  
  // Three trivial cases:
  if ( isZero() && x.isZero() )
    return   EQUAL_TO;
  
  if ( (sign == positive) && (x.sign == negative) )
    return GREATER_THAN;
  
  if ( (sign == negative) && (x.sign == positive) )
    return    LESS_THAN;
  
  
  /* Non-trivial case:
     Only the combinations (+,+) and (-,-) are possible !
     Procede in two steps:
     1. Assume that the pattern was (+,+).
     2. Switch the result, if the pattern was 
        originally (-,-).
  */

  // ad (1):
  ComparisonType temporaryComp;
 
  // Two trivial cases:
  if (size() > x.size())
    temporaryComp = GREATER_THAN;
  else if (size() < x.size())
    temporaryComp = LESS_THAN;

  // Non-trivial case:
  else { 

    list<Word>::const_reverse_iterator first1 =   rbegin();
    list<Word>::const_reverse_iterator  last1 =   rend  ();
    list<Word>::const_reverse_iterator first2 = x.rbegin();
    
    pair<list<Word>::const_reverse_iterator,
         list<Word>::const_reverse_iterator >
      result =  mismatch(first1, last1, first2);

    // No difference found.
    if (result.first == rend())
      return EQUAL_TO;

    // Difference !
    if (*result.first > *result.second)
      temporaryComp = GREATER_THAN;
    else
      temporaryComp = LESS_THAN;

  } // else
    
  // ad (2):
  if ( (sign == negative) && (x.sign == negative) ) {
    if (temporaryComp == GREATER_THAN)
      return LESS_THAN;
    if (temporaryComp == LESS_THAN)
      return GREATER_THAN;
  }

  return temporaryComp;
} // ComparisonType BigInt::compare(


bool BigInt::operator==(const BigInt& x) const {
  return compare(x)&EQUAL_TO;
}


bool BigInt::operator!=(const BigInt& x) const {
  return compare(x)&(LESS_THAN|GREATER_THAN);
}


bool BigInt::operator< (const BigInt& x) const {
  return compare(x)&LESS_THAN;
}


bool BigInt::operator<=(const BigInt& x) const {
  return compare(x)&(LESS_THAN|EQUAL_TO);
}


bool BigInt::operator> (const BigInt& x) const {
  return compare(x)&GREATER_THAN;
}


bool BigInt::operator>=(const BigInt& x) const {
  return compare(x)&(EQUAL_TO|GREATER_THAN);
}


void BigInt::removeLeadingZeros() {

  // Nothing to do.
  if (size()<2)
    return;
  
  list<Word>::iterator i = end();
  while (--i != begin()) {
    if (*i == 0) {
      wordList.erase(i);
    }
    else {
      break;
    }
  } // while (--i != begin())

} // void BigInt::removeLeadingZeros()


BigInt BigInt::operator*(Word x) const {

  assert(size() >= 1);

  if ( (x==0) || isZero() )
    return (BigInt)0;

  list<Word>::const_iterator thisIterator = wordList.begin();
  BigInt result;

  Word upperOld     ;
  Word upperNew = 0 ;
  Word lowerOld     ;
  Word lowerNew = 0 ;
  Word carry    = 0 ;

  wordMultiply(x,*thisIterator,upperOld,lowerOld);
  result.wordList.push_back(lowerOld); 

  while(++thisIterator != end()) {    

    wordMultiply(x,*thisIterator,upperNew,lowerNew);
    result.wordList.push_back(
			      wordAdd(upperOld,
				      lowerNew,
				      carry   ,
				      carry
				      )
			      );

    // Update
    lowerOld = lowerNew;
    upperOld = upperNew;
  } // while(++thisIterator != end())
 
  if ( (upperOld+carry) != 0) {
    result.wordList.push_back(upperOld+carry);
  }

  return result;

} // BigInt BigInt::operator*(Word x) const


BigInt BigInt::operator*(const BigInt& x) const {
  assert (size()>=1 && x.size()>=1);

  if (isZero() || x.isZero())
    return (BigInt)0;

  list<Word>::const_reverse_iterator thisIterator = rbegin();
  BigInt sum( x * (*thisIterator++) );

  for (; thisIterator != rend(); thisIterator++) {
    sum.wordList.push_front(0);
    sum += x * (*thisIterator);
  }

  sum.sign = multDivSignRule(sign,x.sign);
  return sum;

} // BigInt BigInt::operator*(const BigInt& x) const


BigInt& BigInt::operator*=(const BigInt& x) {
  BigInt temp(*this * x);
  *this = temp;

  return *this; 
} // BigInt& BigInt::operator*=(const BigInt& x)


// Bitwise paper-and-pencil algorithm.
BigInt BigInt::operator/(const BigInt& y) const {

  // Assert that the denominator is not zero.
  assert(!y.isZero());

  // Special case: Numerator=0.
  if (isZero())
    return (BigInt)0;
  
  BigInt result   (0);
  BigInt remainder(0);
  // Inlined absolute value.
  BigInt yAbs(y>=0 ? y : -y);

  list<Word>::const_reverse_iterator thisIterator = rbegin();

  for( int bit = size()*BITS_PER_WORD - 1; bit >= 0; --bit ) {
    remainder <<= 1;
    result    <<= 1;

    remainder.front() |= 
      ( (*thisIterator >> (bit%BITS_PER_WORD)) & 1 ) ;
     
    if( remainder >= yAbs ) {
      remainder      -= yAbs;
      result.front() |= 1   ;
    }

    if ( (bit%BITS_PER_WORD == 0) && (bit > 0) ) {
      thisIterator++;
    }
  
  } // for( int bit = size()*BITS_PER_WORD - 1; ...

  result.sign = multDivSignRule(sign, y.sign);

  return result;
} // BigInt BigInt::operator/(const BigInt& y) const


BigInt& BigInt::operator/=(const BigInt& y) {
  assert(!y.isZero());

  BigInt temp(*this / y);
  *this = temp;

  return *this; 
} // BigInt& BigInt::operator/=(const BigInt& y)


BigInt BigInt::operator%(const BigInt& y) const {

  // Assert that the denominator is not zero.
  assert(!y.isZero());

  // Special case: Numerator = 0.
  if (isZero())
    return (BigInt)0;

  BigInt remainder(0);
  list<Word>::const_reverse_iterator thisIterator = rbegin();

  for( int bit = size()*BITS_PER_WORD - 1; bit >= 0; --bit ) {
    remainder <<= 1;
    remainder.front() |= 
      ( (*thisIterator >> (bit%BITS_PER_WORD)) & 1 );
     
    if (remainder >= y) {
      remainder -= y;
    }
    if ( (bit%BITS_PER_WORD == 0) && (bit>0) ) {
      thisIterator++;
    }

  } // for( int bit = size()*BITS_PER_WORD - 1; ...

  return remainder;
} // BigInt BigInt::operator%(const BigInt& y) const


BigInt& BigInt::operator%=(const BigInt& y) {
  BigInt temp(*this % y);
  *this = temp;

  return *this; 
} // BigInt& BigInt::operator%=(const BigInt& y)


pair<Word,Word> 
BigInt::shiftLeft(Word x, int distance) const {
  assert (distance>=0);

  DoubleWord temp(x << distance);

  Word upperResult = (temp >> BITS_PER_WORD) & WORD_MAX;
  Word lowerResult =  temp                   & WORD_MAX;

  return pair<Word,Word>(upperResult,lowerResult);
} // pair<Word,Word> BigInt::shiftLeft(...


BigInt& BigInt::operator<<=(unsigned long distance) {
  assert(size()>=1);

  // Number of zero words that must be appended on the right.
  unsigned int numberOfZeroWords = distance/BITS_PER_WORD;

  distance %= BITS_PER_WORD;

  if (distance == 0) {
    wordList.insert(begin(),numberOfZeroWords,0);
    return *this;
  }

  pair<Word,Word> shiftResult;
  list<Word>::reverse_iterator thisIterator = rbegin();
  
  shiftResult = shiftLeft(back(),distance);
  back()      = shiftResult.second        ;

  if (shiftResult.first) {
    wordList.push_back(shiftResult.first);
    thisIterator++;
  }

  thisIterator++;

  for (; thisIterator != rend(); thisIterator++) {

    shiftResult = shiftLeft(*thisIterator,distance);
    
    list<Word>::reverse_iterator thisIterator2(thisIterator);
    thisIterator2--;
    *thisIterator2 |= shiftResult. first;
    *thisIterator   = shiftResult.second;
  } // for (; ...

  wordList.insert(begin(),numberOfZeroWords,0);

  return *this;
} // BigInt& BigInt::operator<<=(...


BigInt BigInt::operator<<(unsigned long distance) const {
  BigInt temp(*this);
  temp <<= distance;

  return temp;
} // BigInt BigInt::operator<<(...


pair<Word,Word> 
BigInt::shiftRight(Word x, int distance) const {
  assert(distance>=0);

  DoubleWord temp;
  temp = ( x <<  BITS_PER_WORD );
  temp       >>= distance       ;

  Word upperResult = (temp >> BITS_PER_WORD) & WORD_MAX;
  Word lowerResult =  temp                   & WORD_MAX;

  return pair<Word,Word>(upperResult,lowerResult);
} // pair<Word,Word> BigInt::shiftRight(...


BigInt& BigInt::operator>>=(unsigned long distance) {
  assert(size() > 0);
  
  // Number of zero words that must be erased on the right.
  unsigned int wordsToBeErased = distance/BITS_PER_WORD;
  
  distance %= BITS_PER_WORD;

  while ( (size()>0) && (wordsToBeErased>0) ) {
    wordList.pop_front();
    wordsToBeErased--;
  }

  if (size()==0) {
    wordList.push_front(Word(0));
  }
  else {
    pair<Word,Word> shiftResult;
    list<Word>::iterator thisIterator = begin();

    shiftResult   = shiftRight(*thisIterator,distance);
    *thisIterator = shiftResult.first; 
    thisIterator++;

    for (; thisIterator != end(); thisIterator++) {
      shiftResult = shiftRight(*thisIterator,distance);
 
      list<Word>::iterator thisIterator2(thisIterator);      
      thisIterator2--;
      
      *thisIterator2 |= shiftResult.second;
      *thisIterator   = shiftResult. first;
    } // for (; ... 
    
    removeLeadingZeros();

  } // else

  return *this;
} // BigInt& BigInt::operator>>=(...


BigInt BigInt::operator>>(unsigned long distance) const {
  BigInt temp(*this);
  temp >>= distance;
  return temp;
} // BigInt BigInt::operator>>(...


string BigInt::toString() const {
  BigInt remainder(*this);
  string str;

  if (remainder == 0) {
    str = "0";
  }
  else if (remainder < 0) {
    remainder.negate();
    str = "-"+remainder.toString();
    remainder.negate();
  }
  else { // remainder > 0
    while (remainder > 0) {
      str.insert(str.begin(),(remainder%10).front()+'0');
      remainder /= 10;
    } // while (remainder > 0)
  }

  return str;
} // string BigInt::toString() const
  

// I/O.

ostream& operator<<(ostream& out, const BigInt& toBePrinted) {
  out << toBePrinted.toString();
  return out;
} // ostream& operator<<(...


istream& operator>>(istream& in, BigInt& toBeRead) {
  string inString;
  in >> inString;
  toBeRead = inString;
  return in;
} // istream& operator>>(

