/*******************************************************************/
/*                                                                 */
/* Module           : bigint.hh                                    */
/* Description      : Header file for class BigInt, a class for    */
/*                    arbitrarily large integers.                  */
/* Author           : Sebastian Boenisch                           */
/* Email            : Sebastian.Boenisch@iwr.uni-heidelberg.de     */
/* Copyright        : University of Heidelberg                     */
/* Created on       : (see RCS entry below)                        */
/* $Id: bigint.hh,v 1.8 2002/03/05 17:45:30 boenisch Exp $         */
/* Last modified by :                                              */
/* Last modified on :                                              */
/* Update count     :                                              */
/*                                                                 */
/*******************************************************************/
/*                                                                 */
/*  Date      Name               Changes/Extensions                */
/*  ====      ====               ==================                */
/*                                                                 */
/*******************************************************************/


#ifndef BIGINT_HH
#define BIGINT_HH


// Doxygen mainpage:
/** \mainpage The BigInt Package
 *  \author Sebastian Boenisch, University of Heidelberg
 *  \date   March 20, 2002
 *
 * This documentation of the software package BigInt was
 * automatically generated by the source code documentation
 * tool doxygen. \n\n
 * For further information regarding the installation please 
 * refer to the README file. \n\n
 * For further information regarding the objectives and the
 * contents of this package please refer to the user manual
 * which can be found in LATEX and HTML format under manual/latex
 * and manual/html resp. 
 */


#include <algorithm>
#include <cassert>
#include <climits>
#include <cmath>
#include <cstdlib>
#include <functional>
#include <list>
#include <string>



/* ************************************************** */
/*                      Typedefs                      */ 
/* ************************************************** */

/// General assumption: sizeof(DoubleWord) >= 2*sizeof(Word) !!
typedef unsigned short       Word;

/// General assumption: sizeof(DoubleWord) >= 2*sizeof(Word) !!
typedef unsigned long  DoubleWord;


/* ************************************************** */
/*                 Global constants                   */ 
/* ************************************************** */

/// Assumption: 1 byte = 8 bit
const int BITS_PER_WORD = sizeof(Word)*8;
const int BITS_PER_INT  = sizeof(int) *8;

/// WORD_MAX = 2^BITS_PER_WORD - 1 
const Word WORD_MAX = (1 << BITS_PER_WORD) - 1 ;

/// WORD_MAX_PLUS_ONE = 2^BITS_PER_WORD;
const DoubleWord WORD_MAX_PLUS_ONE = 1 << BITS_PER_WORD ; 

const DoubleWord DOUBLE_WORD_MAX = ULONG_MAX;


/* ************************************************** */
/*                       Enums                        */ 
/* ************************************************** */

/// An enum for comparison of BigInts.
enum ComparisonType {
    EQUAL_TO   = 0x1 ,
  GREATER_THAN = 0x2 ,
     LESS_THAN = 0x4 
};


/// The sign of a BigInt.
enum SignType {positive, negative};



/// A class for arbitrarily large integers.
class BigInt {

private:

  /* ************************************************** */
  /*                      Variables                     */ 
  /* ************************************************** */
  

  /// The sign of a BigInt. 
  /** For some good reasons, this variable was laid out as 
   *  mutable which means that it can be changed even in case 
   *  the BigInt object it belongs to is const. To see that   
   *  this does make sense, take a look at the logic in the
   *  arithmetic member functions operator+=(const BigInt&) 
   *  and operator-(const BigInt&) respectively.
   **/
  mutable SignType sign; 

  /// A doubly linked list that stores the bits of a BigInt.
  list<Word> wordList  ;


public:

  /* ************************************************** */
  /*             Methods (public)                       */ 
  /* ************************************************** */

  // Constructors
  BigInt();
  /// Constructs a BigInt from an integer.
  BigInt(int val);
  BigInt(const list<Word> initList);
  /** Constructs a BigInt from a string. 
  Example: BigInt("1234567890987654321") **/ 
  BigInt(const string& initString );

  // Destructor
  ~BigInt();
 
  // Comparison operators
  bool operator==(const BigInt&) const;
  bool operator!=(const BigInt&) const;
  bool operator< (const BigInt&) const;
  bool operator<=(const BigInt&) const;
  bool operator> (const BigInt&) const;
  bool operator>=(const BigInt&) const;
 

  // Arithmetic functions
  
  const
  BigInt& negate   () const;

  BigInt  operator-() const;

  BigInt& operator+=(const BigInt&)        ; 
  BigInt  operator+ (const BigInt&)  const ;
  
  BigInt& operator-=(const BigInt&)        ;
  BigInt  operator- (const BigInt&)  const ;

  BigInt& operator*=(const BigInt&)        ;
  BigInt  operator* (const BigInt&)  const ;
  
  BigInt& operator/=(const BigInt&)        ;
  BigInt  operator/ (const BigInt&)  const ;
  
  BigInt& operator%=(const BigInt&)        ;
  BigInt  operator% (const BigInt&)  const ;
  
  BigInt& operator<<=(unsigned long)       ;
  BigInt  operator<< (unsigned long) const ;
  
  BigInt& operator>>=(unsigned long)       ;
  BigInt  operator>> (unsigned long) const ;

  BigInt& operator++()   ;
  BigInt& operator++(int);
  BigInt& operator--()   ;
  BigInt& operator--(int);
  

  // Other public functions:

  /// Returns the number of Words.  
  unsigned long  size() const;
  
  /// Returns the sign.
  SignType getSign() const;

  // Type conversion
  string toString() const ;
  float  toFloat () const ;
  double toDouble() const ;

  /// Returns true iff *this is equal to zero.
  bool isZero() const;

  /// Returns true iff *this is an even number.
  bool isEven() const;

  /// Returns true iff *this is an odd  number.
  bool isOdd () const;

  /// Returns true iff *this is positive (or equal to zero).
  bool isPositive () const;

  /// Returns true iff *this is negative (or equal to zero).
  bool isNegative () const;

  
private:

  /* ************************************************** */
  /*              Methods (private)                     */ 
  /* ************************************************** */  

  /// A full adder. 
  Word wordAdd     (Word, Word, Word,  Word&) const;
  /// Subtracts two Words.
  Word wordSub     (Word, Word, Word , Word&) const;
  /// Multiplies two Words.
  void wordMultiply(Word, Word, Word&, Word&) const;

  /// Multiplies a BigInt and a Word.
  BigInt operator* (Word) const;

  ///  Left shift.
  pair<Word,Word> shiftLeft (Word, int) const;
  /// Right shift.
  pair<Word,Word> shiftRight(Word, int) const; 

  /// Returns bit i of *this.
  int getBit(unsigned long i) const;
  /// Returns bit i of Word *j of *this.
  int getBit(unsigned int i, list<Word>::const_iterator j) const;

  /// A well-known rule ... ;-)
  SignType multDivSignRule(SignType sign1, SignType sign2) const;

  /// Compares two BigInts.
  ComparisonType compare(const BigInt&) const;

  // Some standard stuff concerning iterators.
  list<Word>::iterator       begin()      ;
  list<Word>::const_iterator begin() const;

  list<Word>::iterator         end()      ;
  list<Word>::const_iterator   end() const;   

  list<Word>::reverse_iterator       rbegin()      ;
  list<Word>::const_reverse_iterator rbegin() const;

  list<Word>::reverse_iterator         rend()      ;
  list<Word>::const_reverse_iterator   rend() const;

        Word& front()      ;  
  const Word& front() const;

        Word& back()       ;  
  const Word& back() const ;
  
  /// Removes all leading zero Words.
  void removeLeadingZeros();


}; // class BigInt



/* ************************************************** */
/*               Non-member functions                 */ 
/* ************************************************** */

/// Output operator.
ostream& operator<<(ostream&, const BigInt&);
/// Input operator.
istream& operator>>(istream&, const BigInt&);



#endif

