/*******************************************************************/
/*                                                                 */
/* Module           : rank.cc                                      */
/* Description      : Implementation file for function computeRank.*/
/* Author           : Sebastian Boenisch                           */
/* Email            : Sebastian.Boenisch@iwr.uni-heidelberg.de     */
/* Copyright        : University of Heidelberg                     */
/* Created on       : (see RCS entry below)                        */
/* $Id: rank.cc,v 1.6 2002/03/05 17:45:21 boenisch Exp $           */
/* Last modified by :                                              */
/* Last modified on :                                              */
/* Update count     :                                              */
/*                                                                 */
/*******************************************************************/
/*                                                                 */
/*  Date      Name               Changes/Extensions                */
/*  ====      ====               ==================                */
/*                                                                 */
/*******************************************************************/


#include <iostream>

#include "bigint.hh"
#include "matrix.hh"



template<class T>
int computeRank (Matrix<T>& matrix) {

  // The rank of the matrix.
  int rank;

  int n = matrix.getSize().first;  // Number of rows. 
  int m = matrix.getSize().second; // Number of columns.


  // While determining the rank of a matrix
  // using Gaussian elimination, it may be 
  // necessary to permute some rows (or alternatively
  // columns). Instead of really performing this 
  // permutations, we use a vector representing
  // the mapping
  // "virtual" indices i ---> "physical" indices
  //                           permutedRows[i].
  vector<int> permutedRows(n);
  for (int i=0; i<n; i++)  permutedRows[i] = i;

  // Counts the number of elimination steps.  
  int eliminationStep = 0;

  // The pivot element.
  T pivot = (T)0;

  /* Common divisor. 
     Divisor = 2^divisorAlpha * divisorBeta.
     This decomposition makes the division faster, as
     x/(2^y) = x >> y .
  */
  long int divisorAlpha =    0;
  T        divisorBeta  = (T)1;


  // Main loop over the matrix columns. 
  for (int pivotColumn=0; pivotColumn<m; ++pivotColumn) {
       
    // Find the pivot = (pivotRow,pivotColumn).
    int pivotRow;

    int i = eliminationStep-1;
    do {
      i++;
      pivotRow = permutedRows[i];
    } while(     (matrix[pivotRow][pivotColumn]==(T)0) 
	      && (i!=n-1)   
	    );

    pivot = matrix[pivotRow][pivotColumn];

    // Found a pivot !
    if (pivot != (T)0 ) {

      pivot = matrix[pivotRow][pivotColumn];

      // There was a row permutation !
      if (eliminationStep != i) {
	permutedRows[i] = permutedRows[eliminationStep];
	permutedRows[eliminationStep] = pivotRow;
      }

      // Elimination required ?
      if ( (pivotColumn != m-1) && (eliminationStep != n-1) ) {
	
	// Eliminate...
#ifdef PRINTLEVEL
	cout << "Elimination " << eliminationStep+1 << endl;
#endif
	// Iterate over all rows "under" the row
	// containing the pivot.
	for (int i=eliminationStep+1; i<n; ++i) {
	  
	  int actualRow = permutedRows[i];

	  // Iterate over all columns "at the right
	  // hand side" of the column containing the
	  // pivot.
	  for (int actualColumn=pivotColumn+1;
	           actualColumn<m; 
	         ++actualColumn   ) {
 
	    T& multiplier = matrix[actualRow][pivotColumn];
	   
	    // Eliminate.
	    matrix[actualRow][actualColumn] = 
	             (pivot * matrix[actualRow][actualColumn]
	       - multiplier * matrix[ pivotRow][actualColumn]);
	    
	    // Division by common divisor.
	    matrix[actualRow][actualColumn] >>= divisorAlpha;
	    matrix[actualRow][actualColumn]  /= divisorBeta ;

	  } // for (int actualColumn=pivotColumn+1;...
	  
	} // for (int i=eliminationStep+1;...
	
	
	for (int i=pivotRow+1; i<n; i++) {
	  matrix[i][pivotColumn] = (T)0;
	}
	
	// Analyze pivot for faster division in next step.
	divisorAlpha = 0;
	while (pivot.isEven()) {
	  pivot >>= 1;
	  divisorAlpha++;
	}
	divisorBeta = pivot;

	// Next elimination step.
	eliminationStep++;
     
      } // if ( (pivotColumn != m-1) && ...
 
    } // if (pivot != (T)0 )

  } // for (int pivotColumn=0;...


  // Determine rank.
  rank = eliminationStep;
  if (pivot != (T)0)
    ++rank;
  
  return rank;
}

