/*******************************************************************/
/*                                                                 */
/* Module           : rank.hh                                      */
/* Description      : Header file for function computeRank, which  */
/*                    computes the rank of a matrix.               */
/* Author           : Sebastian Boenisch                           */
/* Email            : Sebastian.Boenisch@iwr.uni-heidelberg.de     */
/* Copyright        : University of Heidelberg                     */
/* Created on       : (see RCS entry below)                        */
/* $Id: rank.hh,v 1.9 2002/03/05 17:45:30 boenisch Exp $           */
/* Last modified by :                                              */
/* Last modified on :                                              */
/* Update count     :                                              */
/*                                                                 */
/*******************************************************************/
/*                                                                 */
/*  Date      Name               Changes/Extensions                */
/*  ====      ====               ==================                */
/*                                                                 */
/*******************************************************************/


#ifndef RANK_HH
#define RANK_HH


#include "matrix.hh"


/** This function computes the rank of an integer matrix.
 *  The integer type can be specified through the template parameter.
 *  The rank computation uses a variant of direct Gaussian elimination,
 *  which is fitted for the combination integer entries/exact arithmetic.
 *  It avoids the occurence of rational numbers which behavior can not
 *  be expected by standard Gauss elimination. \n
 *  To see how this works, we will do some of the needed manipulations
 *  in the special case of a 3x4 matrix; then you will surely see very quickly
 *  how the cat jumps: \n
 *  \f$ A := A^{(0)} := \left( \begin{array}{cccc}
 *  a_{11} & a_{12} & a_{13} & a_{14} \\
 *  a_{21} & a_{22} & a_{23} & a_{24} \\
 *  a_{31} & a_{32} & a_{33} & a_{34} \\
 *  a_{41} & a_{42} & a_{43} & a_{44} \\
 *  \end{array} \right) \f$
 *  \n 
 *  \f$ A^{(1)} := \left( \begin{array}{cccc}
 *  a_{11} & a_{12} & a_{13} & a_{14} \\
 *  0      & a_{11}a_{22}-a_{21}a_{12} & a_{11}a_{23}-a_{21}a_{13} & a_{11}a_{24}-a_{21}a_{14} \\
 *  0      & a_{11}a_{32}-a_{31}a_{12} & a_{11}a_{33}-a_{31}a_{13} & a_{11}a_{34}-a_{31}a_{14} \\
 *  0      & a_{11}a_{42}-a_{41}a_{12} & a_{11}a_{43}-a_{41}a_{13} & a_{11}a_{44}-a_{41}a_{14} \\
 *  \end{array} \right) \f$
 *  \n \n
 *  \f$ A^{(2)} := \left( \begin{array}{cccc}
 *  a_{11} & a_{12} & a_{13} & a_{14} \\
 *  0      & a_{11}a_{22}-a_{21}a_{12} & a_{11}a_{23}-a_{21}a_{13} & a_{11}a_{24}-a_{21}a_{14} \\
 *  0      & 0      &  A^{(2)}_{33}  & A^{(2)}_{34}  \\
 *  0      & 0      &  A^{(2)}_{43}  & A^{(2)}_{44}  \\
 *  \end{array} \right) \f$
 * \n \n with \n\n
\f$ A^{(2)}_{33} = (a_{11}a_{22}-a_{21}a_{12}) \cdot (a_{11}a_{33}-a_{31}a_{13}) - (a_{11}a_{32}-a_{31}a_{12}) \cdot (a_{11}a_{23}-a_{21}a_{13}) \f$
* \n  
\f$ A^{(2)}_{34} = (a_{11}a_{22}-a_{21}a_{12}) \cdot (a_{11}a_{34}-a_{31}a_{14}) - (a_{11}a_{32}-a_{31}a_{12}) \cdot (a_{11}a_{24}-a_{21}a_{14}) \f$ \n
\f$ A^{(2)}_{43} = ... \f$ \n
\f$ A^{(2)}_{44} = ... \f$ \n
 * \n\n After expanding these products, one can see that it is 
 * possible to divide \f$ A^{(2)}_{33} , A^{(2)}_{34} ,  A^{(2)}_{43} and
 A^{(2)}_{44} \f$
 * by \f$ a_{11} \f$ (without creating any non-integer
 * numbers and, of course, without changing the rank). \n 
 * This \f$ \textbf{division by the pivot element before last} \f$ is 
 * also possible in the remaining elimination steps
 * and leads to a slower increase of the entries
 * during the elimination process.
**/
template<class T>
int computeRank (Matrix<T>& matrix);



/* ************************************************** */
/*             Include implementation                 */ 
/* ************************************************** */

#include "rank.cc"



#endif

