/*******************************************************************/
/*                                                                 */
/* Module           : rational.cc                                  */
/* Description      : Implementation file for class Rational.      */
/*                    rational numbers.                            */
/* Author           : Sebastian Boenisch                           */
/* Email            : Sebastian.Boenisch@iwr.uni-heidelberg.de     */
/* Copyright        : University of Heidelberg                     */
/* Created on       : (see RCS entry below)                        */
/* $Id: rational.cc,v 1.5 2002/03/05 17:45:21 boenisch Exp $       */
/* Last modified by :                                              */
/* Last modified on :                                              */
/* Update count     :                                              */
/*                                                                 */
/*******************************************************************/
/*                                                                 */
/*  Date      Name               Changes/Extensions                */
/*  ====      ====               ==================                */
/*                                                                 */
/*******************************************************************/


#include <algorithm>
#include <cassert>

#include "rational.hh"
#include "utility.hh"



template<class T>
Rational<T>::Rational() 
  : numerator(0), denominator(1) {
}


template<class T>
Rational<T>::Rational(const T& num) 
  : numerator(num), denominator(1) {
}


template<class T>
Rational<T>::Rational(const T& num, const T& denom) 
  : numerator(num), denominator(denom) {
  
  assert( denominator == (T)0 );
  normalize();
} // Rational<T>::Rational(const T& num, const T& denom) 


template<class T>
T Rational<T>::get_numerator() const { 
  return numerator; 
}


template<class T>
T Rational<T>::get_denominator() const { 
  return denominator; 
}


template<class T>
Rational<T> Rational<T>::operator-() {
  return (Rational<T>)(-1) * *this;
}


template<class T>
Rational<T>& Rational<T>::operator+=(const Rational<T>& y) {
  numerator   = numerator * y.denominator + denominator * y.numerator;
  denominator = denominator * y.denominator;
  normalize();
  
  return *this;
} // Rational<T>& Rational<T>::operator+=(const Rational<T>& y)


template<class T>
Rational<T>& Rational<T>::operator-=(const Rational<T>& y) {
  numerator   = numerator * y.denominator - denominator * y.numerator;
  denominator = denominator * y.denominator;
  normalize();
  
  return *this;
} // Rational<T>& Rational<T>::operator-=(const Rational<T>& y)
  

template<class T>
Rational<T>& Rational<T>::operator*=(const Rational<T>& y) {
  numerator *= y.numerator;
  denominator *= y.denominator;
  normalize();

  return *this;
} // Rational<T>& Rational<T>::operator*=(const Rational<T>& y)


template<class T>
Rational<T>& Rational<T>::operator/=(const Rational<T>& y) {
  numerator *= y.denominator;
  denominator *= y.numerator;
  normalize();
  
  return *this;
} // Rational<T>& Rational<T>::operator/=(const Rational<T>& y)


template<class T>
Rational<T>& Rational<T>::frac() {
  assert(numerator != 0);
  swap(numerator,denominator);
  
  return *this;
} // Rational<T>& Rational<T>::frac()


template<class T>
bool Rational<T>::operator==(const Rational<T>& y) const {
  if ( (numerator == y.numerator) && (denominator == y.denominator) )
    return true;
  return false;
} // bool Rational<T>::operator==(const Rational<T>& y) const


template<class T>
bool Rational<T>::operator!=(const Rational<T>& y) const {
  return !(*this == y);
}


template<class T>
bool Rational<T>::operator<(const Rational<T>& y) const {
  if ( (numerator < 0) && (y.numerator >= 0) ) 
    return true;
  if ( (numerator >= 0) && (y.numerator < 0) )
    return false;
  
  T ad = numerator * y.denominator;
  T bc = denominator * y.numerator;
  if (ad < bc)
    return true;
  
  return false;
} // bool Rational<T>::operator<(const Rational<T>& y) const


template<class T>
bool Rational<T>::operator>(const Rational<T>& y) const {
  if ( (numerator >= 0) && (y.numerator < 0) ) 
    return true;
  if ( (numerator < 0) && (y.numerator >= 0) )
    return false;
  
  T ad = numerator * y.denominator;
  T bc = denominator * y.numerator;
  if (ad > bc)
    return true;
  
  return false;
} // bool Rational<T>::operator>(const Rational<T>& y) const


template<class T>
bool Rational<T>::operator<=(const Rational<T>& y) const {
  return !(*this > y);
}


template<class T>
bool Rational<T>::operator>=(const Rational<T>& y) const {
  return !(*this < y);
}


template<class T>
Rational<T>& Rational<T>::normalize() {
  int sign = 1;
  if (numerator < 0) {
    sign = -sign;
    numerator = -numerator;
  }
  if (denominator < 0) {
    sign = -sign;
    denominator = -denominator;
  }
  
  T divisor = gcd(numerator, denominator);
  numerator = sign>=0 ? numerator/divisor : -numerator/divisor;
  denominator = denominator/divisor;
  
  return *this;
} // Rational<T>& Rational<T>::normalize()


template<class T>
double Rational<T>::toDouble() const {
  return numerator.toDouble() / denominator.toDouble();
}



/* ************************************************** */
/*               Non-member functions                 */ 
/* ************************************************** */


template <class T>
Rational<T> operator+(const Rational<T>& x, const Rational<T>& y) {
  Rational<T> temp = x;
  return temp += y;
}


template <class T>
Rational<T> operator-(const Rational<T>& x, const Rational<T>& y) {
  Rational<T> temp = x;
  return temp -= y;
}


template <class T>
Rational<T> operator*(const Rational<T>& x, const Rational<T>& y) {
  Rational<T> temp = x;
  return temp *= y;
}


template <class T>
Rational<T> operator/(const Rational<T>& x, const Rational<T>& y) {
  Rational<T> temp = x;
  return temp /= y;
}



template <class T>
ostream& operator<<(ostream& out, const Rational<T>& x) {
  out << x.get_numerator() << "/" << x.get_denominator();
  return out;
}


template <class T>
istream& operator>>(istream& in, Rational<T>& x) {
  T num, denom;
  in >> num >> denom;
  x = Rational<T>(num,denom);

  return in;
}

