/*******************************************************************/
/*                                                                 */
/* Module           : solve.cc                                     */
/* Description      : Implementation file for functions            */
/*                    LU_Decomposition and                         */
/*                    LU_ForwardBackwardSubstitution               */
/* Author           : Sebastian Boenisch                           */
/* Email            : Sebastian.Boenisch@iwr.uni-heidelberg.de     */
/* Copyright        : University of Heidelberg                     */
/* Created on       : (see RCS entry below)                        */
/* $Id: solve.cc,v 1.1 2002/03/05 17:45:21 boenisch Exp $          */
/* Last modified by :                                              */
/* Last modified on :                                              */
/* Update count     :                                              */
/*                                                                 */
/*******************************************************************/
/*                                                                 */
/*  Date      Name               Changes/Extensions                */
/*  ====      ====               ==================                */
/*                                                                 */
/*******************************************************************/


#include <iostream>
#include <vector>

#include "matrix.hh"


template <class T>     
T LU_Decomposition(Matrix<T>& matrix, vector<int>& perm) {

  int nRows = matrix.getSize().first;
  int nCols = matrix.getSize().second;

  int detSign = 1;

  // Permutation vector.
  for (int k=0; k<nRows; k++) perm[k] = k;

  for (int k=0; k<nRows-1; k++) {  // main loop

#ifdef PRINTLEVEL
    cout << "Elimination " << k+1 << endl;
#endif

      // find the pivot in column k in rows perm[k], 
      // perm[k+1], ..., perm[n-1]
      int pivotCounter = k; 

      T pivotConcurrent = abs(matrix[perm[k]][k]);
      for (int i=k+1; i<nRows; i++) {

        if (abs(matrix[perm[i]][k]) > pivotConcurrent) {
          pivotConcurrent = abs(matrix[perm[i]][k]); 
          pivotCounter    = i;
        }

      } // for (int i=k+1; ...
     
      if (pivotCounter != k) {
	  swap(perm[k], perm[pivotCounter]); 
	  detSign = -detSign; 
      }

      int actualRow = perm[k]; 
      T pivot = matrix[actualRow][k];  // pivot

      // now eliminate column entries logically 
      // below matrix[actualRow][k]
      for (int i = k+1; i < nRows; i++) {

	int eliminateRow = perm[i];

	if (matrix[eliminateRow][k] != static_cast<T>(0) ) {
	  T mult = matrix[eliminateRow][k]/pivot;
	  matrix[eliminateRow][k] = mult;

	  for (int j =k+1; j<nCols; j++) 
	    matrix[eliminateRow][j] -= mult*matrix[actualRow][j];
	} // if (matrix[eliminateRow][k] != ...
	
      } // for (int i = k+1; ...
      
  } // for (int k=...
  
  // return determinant
  T det = (T)1;
  for (int i=0; i<nRows; ++i) {
      det *= matrix[perm[i]][i];
  }
  return (T)detSign * det;

} // LU_Decomposition 



template <class T>
void LU_ForwardBackwardSubstitution
           (Matrix<T>& LU, vector<T>& sol, vector<int>& perm) {
  
  int nRows = LU.getSize().first;
  int nCols = LU.getSize().second;

  // forward substitution for Ly = Pb.
  for (int i=1; i<nRows; i++) { 
    for (int j=0; j<i; j++) {
      sol[perm[i]] -= LU[perm[i]][j]*sol[perm[j]];
    }
  }

  // back substitution for Ux = y
  vector<T> xx(nRows);  // xx stores solution in correct order
  for (int i=nRows-1; i>=0; i--) {
    for (int j=i+1; j<nCols; j++) { 
      sol[perm[i]] -= LU[perm[i]][j]*xx[j];
    }
    xx[i] = sol[perm[i]] / LU[perm[i]][i];
  }

  sol = xx;             // put solution in sol

} //  LU_ForwardBackwardSubstitution
