/*******************************************************************/
/*                                                                 */
/* Module           : solve.hh                                     */
/* Description      : Header file for functions LU_Decomposition   */
/*                    and LU_ForwardBackwardSubstitution which are */
/*                    used to solve linear systems by means of     */
/*                    direct LU decomposition.                     */
/* Author           : Sebastian Boenisch                           */
/* Email            : Sebastian.Boenisch@iwr.uni-heidelberg.de     */
/* Copyright        : University of Heidelberg                     */
/* Created on       : (see RCS entry below)                        */
/* $Id: solve.hh,v 1.1 2002/03/05 17:45:30 boenisch Exp $          */
/* Last modified by :                                              */
/* Last modified on :                                              */
/* Update count     :                                              */
/*                                                                 */
/*******************************************************************/
/*                                                                 */
/*  Date      Name               Changes/Extensions                */
/*  ====      ====               ==================                */
/*                                                                 */
/*******************************************************************/


#ifndef SOLVE_HH
#define SOLVE_HH


#include <vector>

#include "matrix.hh"


/** Computes the \f$ LU \f$ decomposition \f$ LU = PA. \f$
 * This function takes as argument a regular matrix which will be
 * overwritten with the (possibly permuted) \f$ LU \f$ decomposition.
 * Of course, we do not build the permutation matrix \f$ P \f$ explicitly
 * but only the permutation vector perm, which stores the row permutations made
 * during the elimination.  
 * The unpermuted LU decomposition then can be rebuild by using the
 * row permutation vector perm 
 * (cf. function LU_ForwardBackwardSubstitution ). \n\n
 * The template parameter T stands for any type representing integers
 * or rational numbers.
 * \return The function returns the determinant of \f$ A \f$.
*/
template <class T>     
T LU_Decomposition(Matrix<T>& matrix, vector<int>& perm);


/** Solves the linear system \f$ LUx = Pb \Leftrightarrow Ax = b .\f$
 * This function takes as arguments the (possibly permuted) LU
 * decomposition of a matrix (stored in one matrix !) and the 
 * row permutation vector perm (see function LU_Decomposition).
 * It solves then the linear system \f$ LUx=Pb \f$. 
 * The vector sol stores the solution
 * in correct order. \n\n
 * The template parameter T stands for any type representing integers
 * or rational numbers.
*/     
template <class T>
void LU_ForwardBackwardSubstitution
(Matrix<T>& LU, vector<T>& sol, vector<int>& perm);



/* ************************************************** */
/*             Include implementation                 */ 
/* ************************************************** */

#include "solve.cc"


#endif
