/*******************************************************************/
/*                                                                 */
/* Module           : test.cc                                      */
/* Description      : Main program. Performs some tests.           */
/* Author           : Sebastian Boenisch                           */
/* Email            : Sebastian.Boenisch@iwr.uni-heidelberg.de     */
/* Copyright        : University of Heidelberg                     */
/* Created on       : (see RCS entry below)                        */
/* $Id: test.cc,v 1.8 2002/03/05 17:45:21 boenisch Exp $           */
/* Last modified by :                                              */
/* Last modified on :                                              */
/* Update count     :                                              */
/*                                                                 */
/*******************************************************************/
/*                                                                 */
/*  Date      Name               Changes/Extensions                */
/*  ====      ====               ==================                */
/*                                                                 */
/*******************************************************************/


#include <algorithm>
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <fstream>
#include <numeric>
#include <set>
#include <string>
#include <vector>

#include "bigint.hh"
#include "matrix.hh"
#include "rational.hh"
#include "rank.hh"
#include "solve.hh"



/* ***************************************************************** */
/*                           Functions.                              */
/*              (Only needed for the main program.)                  */
/* ***************************************************************** */

/// Converts a string to an int.
int stringToInt(const string& str) {
  int result = 0;
  int sign;
  
  string::const_iterator i = str.begin();
  if (*i == '-') {
    sign = -1;
    i++;
  }
  else if (*i == '+') {
    sign = +1;
    i++;
  }
  else {
    sign = +1;
  }

  for (; i != str.end(); i++) {
    result = result*10 + (*i-'0');
  }

  return sign * result;
} // int stringToInt(const string& str)


/// Returns the Vandermonde matrix associated with a given vector.
template<class T>
Matrix<T> vandermonde(vector<T> col)
{
  int n = col.size();

  Matrix<T> vander(n);

  for (int i=0; i<n; ++i) {
      vander[i][0] = (T)1;
      
      for (int j=1; j<n; ++j) {
	  vander[i][j] = vander[i][j-1] * col[i];
      }
  }

  return vander;

} // Matrix<T> vandermonde(vector<T> col)


/// Returns the famous ill-conditioned Hilbert matrix of dimension n.
template<class T>
Matrix<Rational<T> > hilbert (int n) {
  Matrix<Rational<T> > H(n);

  for (int i=0; i<n; ++i)
    for (int j=0; j<n; ++j)
      H[i][j] = Rational<T>(1,(T)(i+j+1));
   
  return H;

} // Matrix<Rational<T> > hilbert (int n)


/// Returns n! .
BigInt factorial(BigInt n) 
{
  if ( (n==0) || (n==1) )
    return 1;
  
  BigInt result(1);

  for (BigInt k=1; k<=n; k++) {
    result *= k;
  }

  return result;
} // BigInt factorial(BigInt n) 


/// Returns a regular quadratic matrix with random entries. 
/// The elements which do not belong to the main diagonal
/// are guaranteed to be contained in the interval [-range,+range].
/// The diagonal entries may have a greater absolut value. 
template<class T>
Matrix<T> constructRegularRandomMatrix(int n, int range) {

  Matrix<T> result(n);

  // First, fill the matrix with random integer entries 
  // which are guaranteed to be contained in 
  // the interval [-range,+range].
  for (int i=0; i<n; ++i) {   
    for (int j=0; j<n; ++j) {   
      result[i][j] = (T) (rand()%(2*range) - range);
    }
  }

  // Then augment the main diagonal entries using
  // Gerschgorin's theorem to make the matrix regular.
  for (int i=0; i<n; ++i) { 
    T sum = (T)0;
    for (int j=0; j<n; ++j) {
      if (i!=j) {
	sum += abs(result[i][j]);
      }
    }
    result[i][i] = sum+(T)1;
  }

  return result;
} // Matrix<T> constructRegularRandomMatrix(...


/// Returns a random matrix.
/// The elements are guaranteed to be contained in the 
/// interval [-range,+range].
template<class T>
Matrix<T> randomMatrix (int n, int m, int range=10) {
  Matrix<T> result(n,m);

  for (int i=0; i<n; ++i) {
    for (int j=0; j<m; ++j) {
      result[i][j] = (T) (rand()%(2*range) - range);
    }
  }
  
  return result;

} // Matrix<T> randomMatrix (int n, int m, int range=10)


/// Returns a nxm matrix with integer entries which is guaranteed to
/// have the rank specified by the parameter rank. (Provided that
/// rank <= min(n,m) !!) 
template<class T>
Matrix<T> constructRandomMatrix (int n, int m, int rank, int range=10) {
 
  assert(rank <= min(n,m));

  // First assume the case n>m.
  bool transpose = (m > n);
  if (transpose) swap(n,m); 

  // Result
  Matrix<T> result(n,m);

  // Regular matrix of size nxn
  Matrix<T> regular(n);
  regular = constructRegularRandomMatrix<T>(n,range);

  // Take the first rank columns of the regular matrix.
  // Clear the remaining columns.
  for (int i=0; i<n; ++i) {
    for (int j=0; j<m; ++j) {
      if (j<rank)
	result[i][j] = regular[i][j];
      else
	result[i][j] = (T)0;	
    }
  }

  // Fill the zero columns with linear combinations of the
  // first rank columns. 
  for (int j=rank; j<m; ++j) {  
    for (int j2=0; j2<rank; ++j2) {
      int sign = (rand()%2) ? 1 : -1;
      for (int i=0; i<n; ++i) {
	result[i][j] += (T)sign * result[i][j2]; 
      } // for (int i = 0; ...
    } // for (int j2 = 0;...  
  } // for (int j=rank;...


  // Permute some rows to force row permutations 
  // during any Gaussian elimination.

  int i2 = rank-1;
  for (int i=0; i<rank; i+=2) {
    if (++i2 >= n)
      break;
    
    for (int j=0; j<m; ++j) {
      swap(result[i][j],result[i2][j]);
    }
  } // for (int i=0; ...
  
  // Return the matrix; transposed, if necessary.
  if (transpose)
    return result.transpose();
  else
    return result;
} // Matrix<T> constructRandomMatrix 




/* ***************************************************************** */
/*                        Main program.                              */
/* ***************************************************************** */


int main(int argc, char** argv) {
  
  if (argc<4) {
    cerr << "Too few arguments !" 
	 << endl
	 << "Please start the GUI test_gui !"
	 << endl;
    exit(-1);
  }

  // Extract command line arguments.
  string application = argv[1];
  string arg1        = argv[2];
  string arg2        = argv[3];

  // Variables for timer.
  clock_t clockStart ;
  clock_t clockEnd   ;
  double timeConsumption;

  // Demo 1 - Factorial.
  if (application == (string)"factorial") {

    BigInt n1(arg2);
    if (n1 < 1) {
      n1 = 1;
    }

    cout << "Demo 1 - Factorial." << endl;
    cout << "=================" << endl;

    // Start timer:
    clockStart = clock();

    BigInt fac1 = factorial(n1-1) ;   // (n-1)!
    BigInt fac2 = fac1 * n1     ;   //  n!
    BigInt quot = fac2/fac1     ;   // (n-1)!/n!

    cout << endl;

    if (arg1 == (string)"0") {
      cout << (n1-1) << "! = " << endl;    
      cout << fac1   << endl;
    
      cout << endl;

      cout << n1     << "! = " << endl;
      cout << fac2   << endl; 

      cout << endl;
    } // if (arg1 == (string)"0") 
    
    cout << n1   << "! / " << (n1-1) << "! = " << endl;
    cout << quot << endl;

    cout << endl;

    // Stop timer:
    clockEnd = clock();
    timeConsumption = 
      (double)(clockEnd-clockStart)/(double)CLOCKS_PER_SEC ;

    cout << "Time consumption: "
	 << timeConsumption
	 << " seconds."
	 << endl;
    
  } // if (application == (string)"factorial")

  
  // Demo 2 - Determinant.
  if (application == (string)"determinant") {

    cout << "Demo 2 - Determinant." << endl;
    cout << "=====================" << endl;

    // Start timer:
    clockStart = clock();

    // Dimension of matrix.
    int n2( stringToInt(arg2) );

    typedef BigInt                 IntDataType2; 
    typedef Rational<IntDataType2>    DataType2;
    
    Matrix<DataType2> matrix2(n2);

    if (arg1 == (string)"vandermonde") {
    
      // Construct Vandermonde matrix from 
      // the vector (1,1/2,1/3,...)'  .
      vector<DataType2> v2(n2);
      int j=1;
      for (int i=0; i<n2; ++i) {
	v2[i] = DataType2(1,j);
	++j;
      }
      
      matrix2 = vandermonde(v2);
 
      cout << "Determinante Vandermondesche Matrix (n="
	   << n2 << ") = "
	   << endl;
     
    } // if (arg1 == (string)"vandermonde")

    else if (arg1 == (string)"hilbert") {
      matrix2 = hilbert<IntDataType2>(n2);

      cout << "Determinante Hilbert Matrix (n= "
	   << n2 << ") = "
	   << endl;
      
    } // else if (arg1 == (string)"hilbert")

    vector<int> perm2(n2);
    DataType2 det = LU_Decomposition(matrix2,perm2);
 
    cout << det << endl;


    // Stop timer:
    clockEnd = clock();
    timeConsumption = 
      (double)(clockEnd-clockStart)/(double)CLOCKS_PER_SEC ;
    
    cout << "Time consumption: "
	 << timeConsumption
	 << " seconds."
	 << endl;
    

  } // if (application == (string)"determinant")
  

  // Demo 3 - Solve linear systems.
  if (application == (string)"solve") {

    cout << "Demo 3 - Solve linear systems." << endl;
    cout << "==============================" << endl;

    // Start timer:
    clockStart = clock();

    // Dimension of matrix.
    int n3( stringToInt(arg2) );

    typedef BigInt                 IntDataType3; 
    typedef Rational<IntDataType3>    DataType3;
    
    Matrix<DataType3> matrix3(n3);

    if (arg1 == (string)"vandermonde") {      
      // Construct Vandermonde matrix from 
      // the vector (1,2,3,...)'  .

      vector<DataType3> v3(n3);
      int j=1;
      for (int i=0; i<n3; ++i) {
	v3[i] = (DataType3)j;
	++j;
      }
      matrix3 = vandermonde(v3);
     
    } // if (arg1 == (string)"vandermonde")

    else if (arg1 == (string)"hilbert") {
      // Use Hilbert matrix of dimension n3.

      matrix3 = hilbert<IntDataType3>(n3);

    } // else if (arg1 == (string)"hilbert")


    // Solution vector = (1,1,1,...)'
    vector<DataType3> x(n3);
    for (int i=0; i<n3; ++i) {
      x[i] = (DataType3)(IntDataType3)1;
    }

    // Compute right hand side.
    vector<DataType3> b(n3);
    for (int i=0; i<n3; ++i) {
      b[i] = (DataType3)(IntDataType3)0;
      for (int j=0; j<n3; ++j) {
	b[i] += matrix3[i][j]*x[j];
      }
    }

    // Permutation vector for LU decomposition.
    vector<int> perm3(n3);

    // Solve.
    LU_Decomposition              (matrix3,    perm3);
    LU_ForwardBackwardSubstitution(matrix3, b, perm3);
    
    // Print the result. Should be identical with x !
    cout << b << endl;

    // Stop timer:
    clockEnd = clock();
    timeConsumption = 
      (double)(clockEnd-clockStart)/(double)CLOCKS_PER_SEC ;
    
    cout << "Time consumption: "
	 << timeConsumption
	 << " seconds."
	 << endl;
    
  } // if (application == (string)"solve")

    // Demo 4 - Rank.
  if (application == (string)"rank") {

    cout << "Demo 4 - Determine rank." << endl;
    cout << "========================" << endl; 

    // Dimensions of the matrix which is going to be read.
    string N4        = arg1;
    string M4        = arg2;

    int n4     = stringToInt(N4);
    int m4     = stringToInt(M4);

    typedef BigInt DataType4;

    // Open file containing matrix:
    if (argc<5) {
      cerr << "Too few arguments !"
	   << endl
	   << "Please start the GUI test_gui !"
	   << endl;
      exit(-1);
    }
    char* arg3 = argv[4];
    cout << arg3 << endl;

    ifstream matrixFile(arg3) ;
    if ( !matrixFile.is_open() ) {
      cerr << "Couldn't open file. Aborting."
	   << endl;
      exit(-1);
    }

  Matrix<DataType4> matrix4(n4,m4);

  // Read matrix.
  for (int i=0; i<n4; ++i) {
    for(int j=0; j<m4; ++j) {
      
      int entry;
      matrixFile >> entry;
      matrix4[i][j] = (DataType4)entry;

      char k;
      matrixFile >> k;
      if (k != ',')
	exit(1);

    }
  }
  
  matrixFile.close();

  // Start timer:
  clockStart = clock();

  // Compute rank.
  int rank = computeRank(matrix4);  
  cout << "Rank : " << rank << endl;

  // Stop timer:
  clockEnd = clock();
  timeConsumption = 
    (double)(clockEnd-clockStart)/(double)CLOCKS_PER_SEC ;

  cout << "Time consumption: "
       << timeConsumption
       << " seconds."
       << endl;

  } // if (application == (string)"rank")
  

  return 0;
}
